#include <linux/joystick.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <gtk/gtk.h>
#include <gnome.h>
#include <unistd.h>
#include <sys/timeb.h>
#include <string.h>

#define MAX_PUNTI 20

static GtkWidget **assi = NULL;
static GtkWidget **barre = NULL;
static GtkWidget **toggle = NULL;
static GtkWidget **res_ax = NULL;
static GtkWidget **res_bt = NULL;
static int *reset_a = NULL;
static int *reset_b = NULL;
static GtkWidget *abox, *pbox, *tbox, *label, *area, *dial;
static GtkWidget *rabox, *rbbox;
static int fd = 0;
static guint8 na = 0, nb = 0;
static gint tag = 0;
static GdkPixmap *pixmap = NULL;
static int xy[2] = {0, 0};
static int bottom = 0;
static GdkPoint punti1[MAX_PUNTI], punti2[MAX_PUNTI];
static int npunti = 0;
static long diff = 11071982;

static int adj = 6;
static GdkPixmap *cursor;
static const gchar *xpm_cursor[] = {
  "13 13 2 1",
  "  c #FFFFFFFF0000",
  "B c #00000000FFFF",
  "BBBBBBBBBBBBB",
  "BBBBBBBBBBBBB",
  "BB         BB",
  "BB         BB",
  "BB    B    BB",
  "BB    B    BB",
  "BB  BBBBB  BB",
  "BB    B    BB",
  "BB    B    BB",
  "BB         BB",
  "BB         BB",
  "BBBBBBBBBBBBB",
  "BBBBBBBBBBBBB"
};

static int ritardo(struct js_event evento) {
  struct timeb t;
  int res;

  ftime(&t);
  res = evento.time - t.time * 1000 - t.millitm;
  if (diff == 11071982) {
    diff = res;
    res = 0;
  } else {
    res -= diff;
  }
  if (res > 0) {
    diff += res;
  }

  return res;
}  

static gint delete_event(GtkWidget *window, GdkEventAny *e, gpointer data) {
  gtk_main_quit();
  return FALSE;
}

static void draw_joystick() {
  GdkRectangle update_rect;
  GdkGC *gc;
  int tx, ty;
  
  gdk_draw_rectangle(pixmap, area->style->white_gc, TRUE, 0, 0,
    area->allocation.width, area->allocation.height);
  
  tx = punti1[npunti - 1].x;
  ty = punti1[npunti - 1].y;
   
  gc = area->style->black_gc; 

  gdk_draw_pixmap(pixmap, gc, cursor, 0, 0, tx - adj, ty - adj, -1, -1);
  
  if (bottom) {
    gdk_draw_line(pixmap, gc, 0, 0, tx - 2, ty - 2);
    gdk_draw_line(pixmap, gc, area->allocation.width - 1, 0, tx + 2, ty - 2);
    gdk_draw_line(pixmap, gc, 0, area->allocation.height - 1, tx - 2, ty + 2);
    gdk_draw_line(pixmap, gc, area->allocation.width - 1, area->allocation.height - 1, tx + 3, ty + 3);

    gdk_draw_lines(pixmap, gc, punti1, npunti);
  }
  
  update_rect.x = 0;
  update_rect.y = 0;
  update_rect.width = area->allocation.width;
  update_rect.height = area->allocation.height;
    
  gtk_widget_draw(area, &update_rect);
}

static void new_point(int x, int y, int b) {
  int tx, ty;

  tx = (area->allocation.width - adj - adj) * (32767.0 + x) / 65536.0 + adj;
  ty = (area->allocation.height - adj - adj) * (32767.0 + y) / 65536.0 + adj;

  if (!b) {
    npunti = 0;
  } else {
    if (npunti == MAX_PUNTI) {
      npunti--;
      g_memmove(&(punti1[0]), &(punti1[1]), sizeof(punti1[0]) * npunti);
      g_memmove(&(punti2[0]), &(punti2[1]), sizeof(punti2[0]) * npunti);  
    }
  }
  punti1[npunti].x = tx;
  punti1[npunti].y = ty;
  punti2[npunti].x = x;
  punti2[npunti].y = y;
  npunti++;

}

static gint configure_event(GtkWidget *widget, GdkEventConfigure *event) {
  int i;

  if (pixmap) {
    gdk_pixmap_unref(pixmap);
    pixmap = NULL;
  }
  pixmap = gdk_pixmap_new(widget->window, widget->allocation.width, 
    widget->allocation.height, -1);
  
  for (i = 0; i < npunti; i++) {
    punti1[i].x = (widget->allocation.width - adj - adj) * (32767.0 + punti2[i].x) / 65536.0 + adj;
    punti1[i].y = (widget->allocation.height - adj - adj) * (32767.0 + punti2[i].y) / 65536.0 + adj;
  }

  draw_joystick();

  return TRUE;
}

static gint expose_event(GtkWidget *widget, GdkEventExpose *event) {

  gdk_draw_pixmap(widget->window, widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
    pixmap, event->area.x, event->area.y, event->area.x, event->area.y,
    event->area.width, event->area.height);
  
  return FALSE;
}

static void process_event(struct js_event evento) {
  char buf[128];
  
  switch (evento.type & ~JS_EVENT_INIT) {

    case 0x01: {
      gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(toggle[evento.number]),
        evento.value);
      bottom &= ~(1 << evento.number);
      bottom |= (evento.value << evento.number);

      if (evento.type & JS_EVENT_INIT) {
        reset_b[evento.number]++;
        sprintf(buf, "%d", reset_b[evento.number]++); 
        gtk_label_set_text(GTK_LABEL(res_bt[evento.number]), buf);
      }
      break;
    }
    case 0x02: {
      gtk_progress_set_value(GTK_PROGRESS(barre[evento.number]), evento.value);
      if (evento.number < 2) {
        xy[evento.number] = evento.value;
      }
      if (evento.type & JS_EVENT_INIT) {
        reset_a[evento.number]++;
        sprintf(buf, "%d", reset_a[evento.number]++); 
        gtk_label_set_text(GTK_LABEL(res_ax[evento.number]), buf);
      }

      break;
    }
  }

  new_point(xy[0], xy[1], bottom);
}

static void input_func(gpointer data, gint source, GdkInputCondition condition) {
  struct js_event evento;
  int letti, gap;

  letti = read(source, &evento, sizeof(evento));

  if (letti == sizeof(evento)) {
    gap = ritardo(evento);
  
    if (!(evento.type & JS_EVENT_INIT)) {
      gtk_dial_set_value(GTK_DIAL(dial), -gap);
    } else {
      npunti = 0;
    }
    process_event(evento);
    draw_joystick();
  }
}

static void clear_joystick() {
  int i;
  
  if (tag) {
    gdk_input_remove(tag);
    tag = 0;
  }
  if (fd) {
    close(fd);
    fd = 0;
  }
  for (i = 0; i < nb; i++) {
    if (toggle[i]) {
      gtk_widget_destroy(toggle[i]);
      toggle[i] = NULL;
    }
    if (res_bt[i]) {
      gtk_widget_destroy(res_bt[i]);
      res_bt[i] = NULL;
    }
  }
  for (i = 0; i < na; i++) {
    if (assi[i]) {
      gtk_widget_destroy(assi[i]);
      assi[i] = NULL;
    }
    if (barre[i]) {
      gtk_widget_destroy(barre[i]);
      barre[i] = NULL;
    }
    if (res_ax[i]) {
      gtk_widget_destroy(res_ax[i]);
      res_ax[i] = NULL;
    }
  }

  g_free(assi);
  g_free(barre);
  g_free(toggle);
  g_free(res_ax);
  g_free(res_bt);
  g_free(reset_a);
  g_free(reset_b);

  na = 0;
  nb = 0;
  
  gtk_label_set_text(GTK_LABEL(label), "no joystick");
}

static void create_view(int nb, int na) {
  int i;
  char buf[128];

  for (i = 0; i < nb; i++) {
    sprintf(buf, "Bottone %d", i);
    toggle[i] = gtk_toggle_button_new_with_label(buf);
    gtk_box_pack_start(GTK_BOX(tbox), toggle[i], TRUE, TRUE, 0);
    gtk_widget_set_sensitive(GTK_WIDGET(toggle[i]), FALSE);

    res_bt[i] = gtk_label_new("");
    gtk_box_pack_start(GTK_BOX(rbbox), res_bt[i], TRUE, TRUE, 0);
  }
  for (i = 0; i < na; i++) {
    sprintf(buf, "Asse %d:", i);
    assi[i] = gtk_label_new(buf);
    gtk_box_pack_start(GTK_BOX(abox), assi[i], TRUE, TRUE, 0);

    barre[i] = gtk_progress_bar_new();
    gtk_box_pack_start(GTK_BOX(pbox), barre[i], TRUE, TRUE, 0);
    gtk_progress_configure(GTK_PROGRESS(barre[i]), 0, -32767, 32767);
    gtk_progress_set_format_string(GTK_PROGRESS(barre[i]), "%v");
    gtk_progress_set_show_text(GTK_PROGRESS(barre[i]), TRUE);

    res_ax[i] = gtk_label_new("");
    gtk_box_pack_start(GTK_BOX(rabox), res_ax[i], TRUE, TRUE, 0);
  }
}

static void create_joystick(const char *joy_name) {
  char buf[128];
  
  clear_joystick();
  
  fd = open(joy_name, O_RDONLY);
  if (fd) {
    ioctl(fd, JSIOCGAXES, &na);
    ioctl(fd, JSIOCGBUTTONS, &nb);

    assi = g_malloc0(na * sizeof(GtkWidget *));
    barre = g_malloc0(na * sizeof(GtkWidget *));
    toggle = g_malloc0(nb * sizeof(GtkWidget *));
    res_ax = g_malloc0(na * sizeof(GtkWidget *));
    res_bt = g_malloc0(nb * sizeof(GtkWidget *));

    reset_a = g_malloc0(na * sizeof(int));
    reset_b = g_malloc0(nb * sizeof(int));
      
    ioctl(fd, JSIOCGNAME(sizeof(buf)), buf);
    gtk_label_set_text(GTK_LABEL(label), buf);
    create_view(nb, na);
    new_point(0, 0, 0);

    tag = gdk_input_add(fd, GDK_INPUT_READ, input_func, NULL); 
  }
}

static GtkWidget *create_win() {
  GtkWidget *window, *hbox, *vbox, *hhbox;
  GtkAdjustment *gtk_adj;

  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_signal_connect(GTK_OBJECT(window), "delete_event", 
		     GTK_SIGNAL_FUNC(delete_event), NULL);
  gtk_window_set_title(GTK_WINDOW(window), "Joystick tester by AndSoft Inc.");

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(window), vbox);
 
    label = gtk_label_new("joystick name");
    gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

    hbox = gtk_hbox_new(TRUE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);
  
      tbox = gtk_vbox_new(TRUE, 0);
      gtk_box_pack_start(GTK_BOX(hbox), tbox, TRUE, TRUE, 0);

      abox = gtk_vbox_new(TRUE, 0);
      gtk_box_pack_start(GTK_BOX(hbox), abox, TRUE, TRUE, 0);

      pbox = gtk_vbox_new(TRUE, 0);
      gtk_box_pack_start(GTK_BOX(hbox), pbox, TRUE, TRUE, 0);

    hhbox = gtk_hbox_new(TRUE, 0);
    gtk_box_pack_start(GTK_BOX(vbox), hhbox, FALSE, FALSE, 0);

      rbbox = gtk_vbox_new(TRUE, 0);
      gtk_box_pack_start(GTK_BOX(hhbox), rbbox, TRUE, TRUE, 0);

      gtk_adj = (GtkAdjustment *)gtk_adjustment_new(0, 0, 3000, 0, 0, 0);
      dial = gtk_dial_new(gtk_adj);
      gtk_box_pack_start(GTK_BOX(hhbox), dial, TRUE, TRUE, 0);
      gtk_dial_set_view_only(GTK_DIAL(dial), TRUE);
    
      rabox = gtk_vbox_new(TRUE, 0);
      gtk_box_pack_start(GTK_BOX(hhbox), rabox, TRUE, TRUE, 0);

    area = gtk_drawing_area_new();
    gtk_box_pack_start(GTK_BOX(vbox), area, TRUE, TRUE, 0);
    gtk_signal_connect(GTK_OBJECT(area), "expose_event", (GtkSignalFunc)expose_event, NULL);
    gtk_signal_connect(GTK_OBJECT(area), "configure_event", (GtkSignalFunc)configure_event, NULL);

  return window;
} 

int main(int argc, char *argv[]) {
  GtkWidget *main_win;

  gtk_init(&argc, &argv);

  main_win = create_win();

  cursor = gdk_pixmap_create_from_xpm_d(main_win->window, NULL, NULL, (gchar **)xpm_cursor);
  create_joystick("/dev/js0");

  gtk_widget_show_all(main_win);

  gtk_main();

  close(fd);
  return 0;
}







